@@ -2,25 +2,32 @@ try:
     from simplejson import dumps, loads
 except ImportError:  # pragma: no cover
     from json import dumps, loads  # noqa
+from inspect import Arguments
 from itertools import chain, tee
 from mimetypes import guess_type, add_type
 from os.path import splitext
 import logging
 import operator
+import sys
 import types
 
 import six
 
 from webob import (Request as WebObRequest, Response as WebObResponse, exc,
                    acceptparse)
+from webob.multidict import NestedMultiDict
 
-from .compat import urlparse, unquote_plus, izip
+from .compat import urlparse, izip
 from .secure import handle_security
 from .templating import RendererFactory
 from .routing import lookup_controller, NonCanonicalPath
-from .util import _cfg, encode_if_needed
+from .util import _cfg, encode_if_needed, getargspec
 from .middleware.recursive import ForwardRequestException
 
+if six.PY3:
+    from .compat import is_bound_method as ismethod
+else:
+    from inspect import ismethod
 
 # make sure that json is defined in mimetypes
 add_type('application/json', '.json', True)
@@ -31,16 +38,24 @@ logger = logging.getLogger(__name__)
 
 class RoutingState(object):
 
-    def __init__(self, request, response, app, hooks=[], controller=None):
+    def __init__(self, request, response, app, hooks=[], controller=None,
+                 arguments=None):
         self.request = request
         self.response = response
         self.app = app
         self.hooks = hooks
         self.controller = controller
+        self.arguments = arguments
 
 
 class Request(WebObRequest):
-    pass
+
+    def __getattribute__(self, name):
+        try:
+            return WebObRequest.__getattribute__(self, name)
+        except UnicodeDecodeError as e:
+            logger.exception(e)
+            abort(400)
 
 
 class Response(WebObResponse):
@@ -98,7 +113,7 @@ def override_template(template, content_type=None):
         request.pecan['override_content_type'] = content_type
 
 
-def abort(status_code=None, detail='', headers=None, comment=None, **kw):
+def abort(status_code, detail='', headers=None, comment=None, **kw):
     '''
     Raise an HTTP status code, as specified. Useful for returning status
     codes like 401 Unauthorized or 403 Forbidden.
@@ -109,12 +124,24 @@ def abort(status_code=None, detail='', headers=None, comment=None, **kw):
     :param comment: A comment to include in the response.
     '''
 
-    raise exc.status_map[status_code](
-        detail=detail,
-        headers=headers,
-        comment=comment,
-        **kw
-    )
+    # If there is a traceback, we need to catch it for a re-raise
+    try:
+        _, _, traceback = sys.exc_info()
+        webob_exception = exc.status_map[status_code](
+            detail=detail,
+            headers=headers,
+            comment=comment,
+            **kw
+        )
+
+        if six.PY3:
+            raise webob_exception.with_traceback(traceback)
+        else:
+            # Using exec to avoid python 3 parsers from crashing
+            exec('raise webob_exception, None, traceback')
+    finally:
+        # Per the suggestion of the Python docs, delete the traceback object
+        del traceback
 
 
 def redirect(location=None, internal=False, code=None, headers={},
@@ -251,7 +278,8 @@ class PecanBase(object):
             module = __import__(name, fromlist=fromlist)
             kallable = getattr(module, parts[-1])
             msg = "%s does not represent a callable class or function."
-            assert hasattr(kallable, '__call__'), msg % item
+            if not six.callable(kallable):
+                raise TypeError(msg % item)
             return kallable()
 
         raise ImportError('No item named %s' % item)
@@ -326,15 +354,14 @@ class PecanBase(object):
         passed the argument specification for the controller.
         '''
         args = []
+        varargs = []
         kwargs = dict()
-        valid_args = argspec.args[1:]  # pop off `self`
+        valid_args = argspec.args[:]
+        if ismethod(state.controller) or im_self:
+            valid_args.pop(0)  # pop off `self`
         pecan_state = state.request.pecan
 
-        def _decode(x):
-            return unquote_plus(x) if isinstance(x, six.string_types) \
-                else x
-
-        remainder = [_decode(x) for x in remainder]
+        remainder = [x for x in remainder if x]
 
         if im_self is not None:
             args.append(im_self)
@@ -354,7 +381,7 @@ class PecanBase(object):
         if [i for i in remainder if i]:
             if not argspec[1]:
                 abort(404)
-            args.extend(remainder)
+            varargs.extend(remainder)
 
         # get the default positional arguments
         if argspec[3]:
@@ -377,21 +404,22 @@ class PecanBase(object):
                 if name not in argspec[0]:
                     kwargs[encode_if_needed(name)] = value
 
-        return args, kwargs
+        return args, varargs, kwargs
 
     def render(self, template, namespace):
-        renderer = self.renderers.get(
-            self.default_renderer,
-            self.template_path
-        )
         if template == 'json':
             renderer = self.renderers.get('json', self.template_path)
-        if ':' in template:
+        elif ':' in template:
+            renderer_name, template = template.split(':', 1)
             renderer = self.renderers.get(
-                template.split(':')[0],
+                renderer_name,
+                self.template_path
+            )
+        else:
+            renderer = self.renderers.get(
+                self.default_renderer,
                 self.template_path
             )
-            template = template.split(':')[1]
         return renderer.render(template, namespace)
 
     def find_controller(self, state):
@@ -404,7 +432,7 @@ class PecanBase(object):
 
         # store the routing path for the current application to allow hooks to
         # modify it
-        pecan_state['routing_path'] = path = req.encget('PATH_INFO')
+        pecan_state['routing_path'] = path = req.path_info
 
         # handle "on_route" hooks
         self.handle_hooks(self.hooks, 'on_route', state)
@@ -417,13 +445,13 @@ class PecanBase(object):
         if self.guess_content_type_from_ext \
                 and not pecan_state['content_type'] \
                 and '.' in path:
-            new_path, extension = splitext(path)
+            _, extension = splitext(path.rstrip('/'))
 
             # preface with a letter to ensure compat for 2.5
             potential_type = guess_type('x' + extension)[0]
 
-            if potential_type is not None:
-                path = new_path
+            if extension and potential_type is not None:
+                path = ''.join(path.rsplit(extension, 1))
                 pecan_state['extension'] = extension
                 pecan_state['content_type'] = potential_type
 
@@ -492,25 +520,34 @@ class PecanBase(object):
             )
             raise exc.HTTPNotFound
 
-        # handle "before" hooks
-        self.handle_hooks(self.determine_hooks(controller), 'before', state)
-
         # fetch any parameters
         if req.method == 'GET':
-            params = dict(req.GET)
+            params = req.GET
+        elif req.content_type in ('application/json',
+                                  'application/javascript'):
+            try:
+                if not isinstance(req.json, dict):
+                    raise TypeError('%s is not a dict' % req.json)
+                params = NestedMultiDict(req.GET, req.json)
+            except (TypeError, ValueError):
+                params = req.params
         else:
-            params = dict(req.params)
+            params = req.params
 
         # fetch the arguments for the controller
-        args, kwargs = self.get_args(
+        args, varargs, kwargs = self.get_args(
             state,
-            params,
+            params.mixed(),
             remainder,
             cfg['argspec'],
             im_self
         )
+        state.arguments = Arguments(args, varargs, kwargs)
 
-        return controller, args, kwargs
+        # handle "before" hooks
+        self.handle_hooks(self.determine_hooks(controller), 'before', state)
+
+        return controller, args + varargs, kwargs
 
     def invoke_controller(self, controller, args, kwargs, state):
         '''
@@ -522,6 +559,15 @@ class PecanBase(object):
         resp = state.response
         pecan_state = req.pecan
 
+        # If a keyword is supplied via HTTP GET or POST arguments, but the
+        # function signature does not allow it, just drop it (rather than
+        # generating a TypeError).
+        argspec = getargspec(controller)
+        keys = kwargs.keys()
+        for key in keys:
+            if key not in argspec.args and not argspec.keywords:
+                kwargs.pop(key)
+
         # get the result from the controller
         result = controller(*args, **kwargs)
 
@@ -539,9 +585,11 @@ class PecanBase(object):
         template = content_types.get(pecan_state['content_type'])
 
         # check if for controller override of template
-        template = pecan_state.get('override_template', template) or (
-            'json' if self.default_renderer == 'json' else None
-        )
+        template = pecan_state.get('override_template', template)
+        if template is None and cfg['explicit_content_type'] is False:
+            if self.default_renderer == 'json':
+                template = 'json'
+
         pecan_state['content_type'] = pecan_state.get(
             'override_content_type',
             pecan_state['content_type']
@@ -566,33 +614,43 @@ class PecanBase(object):
             resp.text = result
         elif result:
             resp.body = result
-        elif response.status_int == 200:
+
+        if pecan_state['content_type']:
+            # set the content type
+            resp.content_type = pecan_state['content_type']
+
+    def _handle_empty_response_body(self, state):
+        # Enforce HTTP 204 for responses which contain no body
+        if state.response.status_int == 200:
             # If the response is a generator...
-            if isinstance(response.app_iter, types.GeneratorType):
+            if isinstance(state.response.app_iter, types.GeneratorType):
                 # Split the generator into two so we can peek at one of them
                 # and determine if there is any response body content
-                a, b = tee(response.app_iter)
+                a, b = tee(state.response.app_iter)
                 try:
                     next(a)
                 except StopIteration:
                     # If we hit StopIteration, the body is empty
-                    resp.status = 204
+                    state.response.status = 204
                 finally:
-                    resp.app_iter = b
+                    state.response.app_iter = b
             else:
                 text = None
-                if response.charset:
-                    # `response.text` cannot be accessed without a charset
-                    # (because we don't know which encoding to use)
-                    text = response.text
-                if not any((response.body, text)):
-                    resp.status = 204
+                if state.response.charset:
+                    # `response.text` cannot be accessed without a valid
+                    # charset (because we don't know which encoding to use)
+                    try:
+                        text = state.response.text
+                    except UnicodeDecodeError:
+                        # If a valid charset is not specified, don't bother
+                        # trying to guess it (because there's obviously
+                        # content, so we know this shouldn't be a 204)
+                        pass
+                if not any((state.response.body, text)):
+                    state.response.status = 204
 
-        if resp.status_int in (204, 304):
-            resp.content_type = None
-        elif pecan_state['content_type']:
-            # set the content type
-            resp.content_type = pecan_state['content_type']
+        if state.response.status_int in (204, 304):
+            state.response.content_type = None
 
     def __call__(self, environ, start_response):
         '''
@@ -604,6 +662,10 @@ class PecanBase(object):
         req = self.request_cls(environ)
         resp = self.response_cls()
         state = RoutingState(req, resp, self)
+        environ['pecan.locals'] = {
+            'request': req,
+            'response': resp
+        }
         controller = None
 
         # handle the request
@@ -663,6 +725,8 @@ class PecanBase(object):
                 self.determine_hooks(state.controller), 'after', state
             )
 
+        self._handle_empty_response_body(state)
+
         # get the response
         return state.response(environ, start_response)
 
@@ -673,15 +737,23 @@ class ExplicitPecan(PecanBase):
         # When comparing the argspec of the method to GET/POST params,
         # ignore the implicit (req, resp) at the beginning of the function
         # signature
-        signature_error = TypeError(
-            'When `use_context_locals` is `False`, pecan passes an explicit '
-            'reference to the request and response as the first two arguments '
-            'to the controller.\nChange the `%s.%s.%s` signature to accept '
-            'exactly 2 initial arguments (req, resp)' % (
+        if hasattr(state.controller, '__self__'):
+            _repr = '.'.join((
                 state.controller.__self__.__class__.__module__,
                 state.controller.__self__.__class__.__name__,
                 state.controller.__name__
-            )
+            ))
+        else:
+            _repr = '.'.join((
+                state.controller.__module__,
+                state.controller.__name__
+            ))
+
+        signature_error = TypeError(
+            'When `use_context_locals` is `False`, pecan passes an explicit '
+            'reference to the request and response as the first two arguments '
+            'to the controller.\nChange the `%s` signature to accept exactly '
+            '2 initial arguments (req, resp)' % _repr
         )
         try:
             positional = argspec.args[:]
@@ -691,11 +763,17 @@ class ExplicitPecan(PecanBase):
         except IndexError:
             raise signature_error
 
-        args, kwargs = super(ExplicitPecan, self).get_args(
+        args, varargs, kwargs = super(ExplicitPecan, self).get_args(
             state, all_params, remainder, argspec, im_self
         )
-        args = [state.request, state.response] + args
-        return args, kwargs
+
+        if ismethod(state.controller):
+            args = [state.request, state.response] + args
+        else:
+            # generic controllers have an explicit self *first*
+            # (because they're decorated functions, not instance methods)
+            args[1:1] = [state.request, state.response]
+        return args, varargs, kwargs
 
 
 class Pecan(PecanBase):
@@ -747,12 +825,14 @@ class Pecan(PecanBase):
             state.hooks = []
             state.app = self
             state.controller = None
+            state.arguments = None
             return super(Pecan, self).__call__(environ, start_response)
         finally:
             del state.hooks
             del state.request
             del state.response
             del state.controller
+            del state.arguments
             del state.app
 
     def init_context_local(self, local_factory):
@@ -766,6 +846,7 @@ class Pecan(PecanBase):
         state.response = _state.response
         controller, args, kw = super(Pecan, self).find_controller(_state)
         state.controller = controller
+        state.arguments = _state.arguments
         return controller, args, kw
 
     def handle_hooks(self, hooks, *args, **kw):
