@@ -1,22 +1,23 @@
+# -*- coding: utf-8 -*-
+
 import sys
 import os
 import json
+import traceback
 import warnings
-if sys.version_info < (2, 7):
-    import unittest2 as unittest  # pragma: nocover
-else:
-    import unittest  # pragma: nocover
 
 import webob
 from webob.exc import HTTPNotFound
+import mock
 from webtest import TestApp
 import six
 from six import b as b_
+from six import u as u_
 from six.moves import cStringIO as StringIO
 
 from pecan import (
     Pecan, Request, Response, expose, request, response, redirect,
-    abort, make_app, override_template, render
+    abort, make_app, override_template, render, route
 )
 from pecan.templating import (
     _builtin_renderers as builtin_renderers, error_formatters
@@ -24,6 +25,11 @@ from pecan.templating import (
 from pecan.decorators import accept_noncanonical
 from pecan.tests import PecanTestCase
 
+if sys.version_info < (2, 7):
+    import unittest2 as unittest  # pragma: nocover
+else:
+    import unittest  # pragma: nocover
+
 
 class SampleRootController(object):
     pass
@@ -68,6 +74,10 @@ class TestEmptyContent(PecanTestCase):
             def explicit_json_body(self):
                 response.json_body = {'foo': 'bar'}
 
+            @expose()
+            def non_unicode(self):
+                return chr(0xc0)
+
         return TestApp(Pecan(RootController()))
 
     def test_empty_index(self):
@@ -77,6 +87,10 @@ class TestEmptyContent(PecanTestCase):
         self.assertEqual(r.headers['Content-Length'], '0')
         self.assertEqual(len(r.body), 0)
 
+    def test_index_with_non_unicode(self):
+        r = self.app_.get('/non_unicode/')
+        self.assertEqual(r.status_int, 200)
+
     def test_explicit_body(self):
         r = self.app_.get('/explicit_body/')
         self.assertEqual(r.status_int, 200)
@@ -137,6 +151,23 @@ class TestAppIterFile(PecanTestCase):
         assert len(r.body) == 0
 
 
+class TestInvalidURLEncoding(PecanTestCase):
+
+    @property
+    def app_(self):
+        class RootController(object):
+
+            @expose()
+            def _route(self, args, request):
+                assert request.path
+
+        return TestApp(Pecan(RootController()))
+
+    def test_rest_with_non_utf_8_body(self):
+        r = self.app_.get('/%aa/', expect_errors=True)
+        assert r.status_int == 400
+
+
 class TestIndexRouting(PecanTestCase):
 
     @property
@@ -231,6 +262,35 @@ class TestObjectDispatch(PecanTestCase):
         assert r.body == b_('/sub/sub/deeper')
 
 
+@unittest.skipIf(not six.PY3, "tests are Python3 specific")
+class TestUnicodePathSegments(PecanTestCase):
+
+    def test_unicode_methods(self):
+        class RootController(object):
+            pass
+        setattr(RootController, '🌰', expose()(lambda self: 'Hello, World!'))
+        app = TestApp(Pecan(RootController()))
+
+        resp = app.get('/%F0%9F%8C%B0/')
+        assert resp.status_int == 200
+        assert resp.body == b_('Hello, World!')
+
+    def test_unicode_child(self):
+        class ChildController(object):
+            @expose()
+            def index(self):
+                return 'Hello, World!'
+
+        class RootController(object):
+            pass
+        setattr(RootController, '🌰', ChildController())
+        app = TestApp(Pecan(RootController()))
+
+        resp = app.get('/%F0%9F%8C%B0/')
+        assert resp.status_int == 200
+        assert resp.body == b_('Hello, World!')
+
+
 class TestLookups(PecanTestCase):
 
     @property
@@ -368,6 +428,11 @@ class TestControllerArguments(PecanTestCase):
                     ', '.join(list(args) + data)
                 )
 
+            @staticmethod
+            @expose()
+            def static(id):
+                return "id is %s" % id
+
             @expose()
             def _route(self, args, request):
                 if hasattr(self, args[0]):
@@ -398,6 +463,16 @@ class TestControllerArguments(PecanTestCase):
         assert r.status_int == 200
         assert r.body == b_('index: This is a test!')
 
+    def test_single_argument_with_plus(self):
+        r = self.app_.get('/foo+bar')
+        assert r.status_int == 200
+        assert r.body == b_('index: foo+bar')
+
+    def test_single_argument_with_encoded_plus(self):
+        r = self.app_.get('/foo%2Bbar')
+        assert r.status_int == 200
+        assert r.body == b_('index: foo+bar')
+
     def test_two_arguments(self):
         r = self.app_.get('/1/dummy', status=404)
         assert r.status_int == 404
@@ -412,6 +487,16 @@ class TestControllerArguments(PecanTestCase):
         assert r.status_int == 200
         assert r.body == b_('index: This is a test!')
 
+    def test_keyword_argument_with_plus(self):
+        r = self.app_.get('/?id=foo+bar')
+        assert r.status_int == 200
+        assert r.body == b_('index: foo bar')
+
+    def test_keyword_argument_with_encoded_plus(self):
+        r = self.app_.get('/?id=foo%2Bbar')
+        assert r.status_int == 200
+        assert r.body == b_('index: foo+bar')
+
     def test_argument_and_keyword_argument(self):
         r = self.app_.get('/3?id=three')
         assert r.status_int == 200
@@ -432,6 +517,16 @@ class TestControllerArguments(PecanTestCase):
         assert r.status_int == 200
         assert r.body == b_('index: 4')
 
+    def test_explicit_json_kwargs(self):
+        r = self.app_.post_json('/', {'id': '4'})
+        assert r.status_int == 200
+        assert r.body == b_('index: 4')
+
+    def test_path_with_explicit_json_kwargs(self):
+        r = self.app_.post_json('/4', {'id': 'four'})
+        assert r.status_int == 200
+        assert r.body == b_('index: 4')
+
     def test_multiple_kwargs(self):
         r = self.app_.get('/?id=5&dummy=dummy')
         assert r.status_int == 200
@@ -442,6 +537,11 @@ class TestControllerArguments(PecanTestCase):
         assert r.status_int == 200
         assert r.body == b_('index: 6')
 
+    def test_json_kwargs_from_root(self):
+        r = self.app_.post_json('/', {'id': '6', 'dummy': 'dummy'})
+        assert r.status_int == 200
+        assert r.body == b_('index: 6')
+
         # multiple args
 
     def test_multiple_positional_arguments(self):
@@ -469,6 +569,11 @@ class TestControllerArguments(PecanTestCase):
         assert r.status_int == 200
         assert r.body == b_('multiple: five, six')
 
+    def test_positional_args_with_json_kwargs(self):
+        r = self.app_.post_json('/multiple', {'one': 'five', 'two': 'six'})
+        assert r.status_int == 200
+        assert r.body == b_('multiple: five, six')
+
     def test_positional_args_with_url_encoded_dictionary_kwargs(self):
         r = self.app_.post('/multiple', {'one': 'Five%20', 'two': 'Six%20%21'})
         assert r.status_int == 200
@@ -519,6 +624,11 @@ class TestControllerArguments(PecanTestCase):
         assert r.status_int == 200
         assert r.body == b_('optional: 4')
 
+    def test_optional_arg_with_json_kwargs(self):
+        r = self.app_.post_json('/optional', {'id': '4'})
+        assert r.status_int == 200
+        assert r.body == b_('optional: 4')
+
     def test_optional_arg_with_url_encoded_kwargs(self):
         r = self.app_.post('/optional', {'id': 'Some%20Number'})
         assert r.status_int == 200
@@ -529,6 +639,11 @@ class TestControllerArguments(PecanTestCase):
         assert r.status_int == 200
         assert r.body == b_('optional: 5')
 
+    def test_multiple_positional_arguments_with_json_kwargs(self):
+        r = self.app_.post_json('/optional/5', {'id': 'five'})
+        assert r.status_int == 200
+        assert r.body == b_('optional: 5')
+
     def test_multiple_positional_url_encoded_arguments_with_kwargs(self):
         r = self.app_.post('/optional/Some%20Number', {'id': 'five'})
         assert r.status_int == 200
@@ -549,6 +664,11 @@ class TestControllerArguments(PecanTestCase):
         assert r.status_int == 200
         assert r.body == b_('optional: 7')
 
+    def test_optional_arg_with_multiple_json_kwargs(self):
+        r = self.app_.post_json('/optional', {'id': '7', 'dummy': 'dummy'})
+        assert r.status_int == 200
+        assert r.body == b_('optional: 7')
+
     def test_optional_arg_with_multiple_url_encoded_dictionary_kwargs(self):
         r = self.app_.post('/optional', {
             'id': 'Some%20Number',
@@ -613,6 +733,11 @@ class TestControllerArguments(PecanTestCase):
         assert r.status_int == 200
         assert r.body == b_('multiple_optional: 1, None, None')
 
+    def test_multiple_optional_positional_args_with_json_kwargs(self):
+        r = self.app_.post_json('/multiple_optional', {'one': '1'})
+        assert r.status_int == 200
+        assert r.body == b_('multiple_optional: 1, None, None')
+
     def test_multiple_optional_positional_args_with_encoded_dict_kwargs(self):
         r = self.app_.post('/multiple_optional', {'one': 'One%21'})
         assert r.status_int == 200
@@ -623,6 +748,11 @@ class TestControllerArguments(PecanTestCase):
         assert r.status_int == 200
         assert r.body == b_('multiple_optional: 1, None, None')
 
+    def test_multiple_optional_positional_args_and_json_kwargs(self):
+        r = self.app_.post_json('/multiple_optional/1', {'one': 'one'})
+        assert r.status_int == 200
+        assert r.body == b_('multiple_optional: 1, None, None')
+
     def test_multiple_optional_encoded_positional_args_and_dict_kwargs(self):
         r = self.app_.post('/multiple_optional/One%21', {'one': 'one'})
         assert r.status_int == 200
@@ -648,6 +778,14 @@ class TestControllerArguments(PecanTestCase):
         assert r.status_int == 200
         assert r.body == b_('multiple_optional: 1, 2, 3')
 
+    def test_multiple_optional_args_with_multiple_json_kwargs(self):
+        r = self.app_.post_json(
+            '/multiple_optional',
+            {'one': '1', 'two': '2', 'three': '3', 'four': '4'}
+        )
+        assert r.status_int == 200
+        assert r.body == b_('multiple_optional: 1, 2, 3')
+
     def test_multiple_optional_args_with_multiple_encoded_dict_kwargs(self):
         r = self.app_.post(
             '/multiple_optional',
@@ -701,6 +839,14 @@ class TestControllerArguments(PecanTestCase):
         assert r.status_int == 200
         assert r.body == b_('variable_args: ')
 
+    def test_variable_args_with_json_kwargs(self):
+        r = self.app_.post_json(
+            '/variable_args',
+            {'id': '3', 'dummy': 'dummy'}
+        )
+        assert r.status_int == 200
+        assert r.body == b_('variable_args: ')
+
     def test_variable_kwargs(self):
         r = self.app_.get('/variable_kwargs')
         assert r.status_int == 200
@@ -727,6 +873,14 @@ class TestControllerArguments(PecanTestCase):
         assert r.status_int == 200
         assert r.body == b_('variable_kwargs: dummy=dummy, id=3')
 
+    def test_multiple_variable_kwargs_with_json_kwargs(self):
+        r = self.app_.post_json(
+            '/variable_kwargs',
+            {'id': '3', 'dummy': 'dummy'}
+        )
+        assert r.status_int == 200
+        assert r.body == b_('variable_kwargs: dummy=dummy, id=3')
+
     def test_multiple_variable_kwargs_with_encoded_dict_kwargs(self):
         r = self.app_.post(
             '/variable_kwargs',
@@ -771,6 +925,14 @@ class TestControllerArguments(PecanTestCase):
         assert r.status_int == 200
         assert r.body == b_('variable_all: 6, day=12, month=1')
 
+    def test_variable_post_with_json_kwargs(self):
+        r = self.app_.post_json(
+            '/variable_all/6',
+            {'month': '1', 'day': '12'}
+        )
+        assert r.status_int == 200
+        assert r.body == b_('variable_all: 6, day=12, month=1')
+
     def test_variable_post_mixed(self):
         r = self.app_.post(
             '/variable_all/7',
@@ -779,6 +941,46 @@ class TestControllerArguments(PecanTestCase):
         assert r.status_int == 200
         assert r.body == b_('variable_all: 7, day=12, id=seven, month=1')
 
+    def test_variable_post_mixed_with_json(self):
+        r = self.app_.post_json(
+            '/variable_all/7',
+            {'id': 'seven', 'month': '1', 'day': '12'}
+        )
+        assert r.status_int == 200
+        assert r.body == b_('variable_all: 7, day=12, id=seven, month=1')
+
+    def test_duplicate_query_parameters_GET(self):
+        r = self.app_.get('/variable_kwargs?list=1&list=2')
+        l = [u_('1'), u_('2')]
+        assert r.status_int == 200
+        assert r.body == b_('variable_kwargs: list=%s' % l)
+
+    def test_duplicate_query_parameters_POST(self):
+        r = self.app_.post('/variable_kwargs',
+                           {'list': ['1', '2']})
+        l = [u_('1'), u_('2')]
+        assert r.status_int == 200
+        assert r.body == b_('variable_kwargs: list=%s' % l)
+
+    def test_duplicate_query_parameters_POST_mixed(self):
+        r = self.app_.post('/variable_kwargs?list=1&list=2',
+                           {'list': ['3', '4']})
+        l = [u_('1'), u_('2'), u_('3'), u_('4')]
+        assert r.status_int == 200
+        assert r.body == b_('variable_kwargs: list=%s' % l)
+
+    def test_duplicate_query_parameters_POST_mixed_json(self):
+        r = self.app_.post('/variable_kwargs?list=1&list=2',
+                           {'list': 3})
+        l = [u_('1'), u_('2'), u_('3')]
+        assert r.status_int == 200
+        assert r.body == b_('variable_kwargs: list=%s' % l)
+
+    def test_staticmethod(self):
+        r = self.app_.get('/static/foobar')
+        assert r.status_int == 200
+        assert r.body == b_('id is foobar')
+
     def test_no_remainder(self):
         try:
             r = self.app_.get('/eater')
@@ -835,6 +1037,16 @@ class TestControllerArguments(PecanTestCase):
         assert r.status_int == 200
         assert r.body == b_('eater: 9, None, day=12, month=1')
 
+    def test_post_empty_remainder_with_json_kwargs(self):
+        r = self.app_.post_json('/eater/9/', {'month': '1', 'day': '12'})
+        assert r.status_int == 200
+        assert r.body == b_('eater: 9, None, day=12, month=1')
+
+    def test_post_remainder_with_json_kwargs(self):
+        r = self.app_.post_json('/eater/9', {'month': '1', 'day': '12'})
+        assert r.status_int == 200
+        assert r.body == b_('eater: 9, None, day=12, month=1')
+
     def test_post_many_remainders_with_many_kwargs(self):
         r = self.app_.post(
             '/eater/10',
@@ -843,6 +1055,14 @@ class TestControllerArguments(PecanTestCase):
         assert r.status_int == 200
         assert r.body == b_('eater: 10, dummy, day=12, month=1')
 
+    def test_post_many_remainders_with_many_json_kwargs(self):
+        r = self.app_.post_json(
+            '/eater/10',
+            {'id': 'ten', 'month': '1', 'day': '12', 'dummy': 'dummy'}
+        )
+        assert r.status_int == 200
+        assert r.body == b_('eater: 10, dummy, day=12, month=1')
+
 
 class TestDefaultErrorRendering(PecanTestCase):
 
@@ -902,6 +1122,20 @@ class TestAbort(PecanTestCase):
         r = app.get('/', status=401)
         assert r.status_int == 401
 
+    def test_abort_keeps_traceback(self):
+        last_exc, last_traceback = None, None
+
+        try:
+            try:
+                raise Exception('Bottom Exception')
+            except:
+                abort(404)
+        except Exception:
+            last_exc, _, last_traceback = sys.exc_info()
+
+        assert last_exc is HTTPNotFound
+        assert 'Bottom Exception' in traceback.format_tb(last_traceback)[-1]
+
 
 class TestScriptName(PecanTestCase):
 
@@ -1167,24 +1401,28 @@ class TestFileTypeExtensions(PecanTestCase):
         return TestApp(Pecan(RootController()))
 
     def test_html_extension(self):
-        r = self.app_.get('/index.html')
-        assert r.status_int == 200
-        assert r.body == b_('.html')
+        for path in ('/index.html', '/index.html/'):
+            r = self.app_.get(path)
+            assert r.status_int == 200
+            assert r.body == b_('.html')
 
     def test_image_extension(self):
-        r = self.app_.get('/image.png')
-        assert r.status_int == 200
-        assert r.body == b_('.png')
+        for path in ('/index.png', '/index.png/'):
+            r = self.app_.get(path)
+            assert r.status_int == 200
+            assert r.body == b_('.png')
 
     def test_hidden_file(self):
-        r = self.app_.get('/.vimrc')
-        assert r.status_int == 204
-        assert r.body == b_('')
+        for path in ('/.vimrc', '/.vimrc/'):
+            r = self.app_.get(path)
+            assert r.status_int == 204
+            assert r.body == b_('')
 
     def test_multi_dot_extension(self):
-        r = self.app_.get('/gradient.min.js')
-        assert r.status_int == 200
-        assert r.body == b_('.js')
+        for path in ('/gradient.min.js', '/gradient.min.js/'):
+            r = self.app_.get(path)
+            assert r.status_int == 200
+            assert r.body == b_('.js')
 
     def test_bad_content_type(self):
         class RootController(object):
@@ -1235,6 +1473,39 @@ class TestFileTypeExtensions(PecanTestCase):
         assert r.status_int == 200
         assert r.body == b_('SOME VALUE')
 
+    def test_content_type_guessing_disabled(self):
+
+        class ResourceController(object):
+
+            def __init__(self, name):
+                self.name = name
+                assert self.name == 'file.html'
+
+            @expose('json')
+            def index(self):
+                return dict(name=self.name)
+
+        class RootController(object):
+
+            @expose()
+            def _lookup(self, name, *remainder):
+                return ResourceController(name), remainder
+
+        app = TestApp(
+            Pecan(RootController(), guess_content_type_from_ext=False)
+        )
+
+        r = app.get('/file.html/')
+        assert r.status_int == 200
+        result = dict(json.loads(r.body.decode()))
+        assert result == {'name': 'file.html'}
+
+        r = app.get('/file.html')
+        assert r.status_int == 302
+        r = r.follow()
+        result = dict(json.loads(r.body.decode()))
+        assert result == {'name': 'file.html'}
+
 
 class TestContentTypeByAcceptHeaders(PecanTestCase):
 
@@ -1688,6 +1959,18 @@ class TestEngines(PecanTestCase):
         result = dict(json.loads(r.body.decode()))
         assert result == {'name': 'Bill'}
 
+    def test_default_json_renderer_with_explicit_content_type(self):
+
+        class RootController(object):
+            @expose(content_type='text/plain')
+            def index(self, name='Bill'):
+                return name
+
+        app = TestApp(Pecan(RootController(), default_renderer='json'))
+        r = app.get('/')
+        assert r.status_int == 200
+        assert r.body == b_("Bill")
+
 
 class TestDeprecatedRouteMethod(PecanTestCase):
 
@@ -1711,3 +1994,271 @@ class TestDeprecatedRouteMethod(PecanTestCase):
             r = self.app_.get('/foo/bar/')
             assert r.status_int == 200
             assert b_('foo, bar') in r.body
+
+
+class TestExplicitRoute(PecanTestCase):
+
+    def test_alternate_route(self):
+
+        class RootController(object):
+
+            @expose(route='some-path')
+            def some_path(self):
+                return 'Hello, World!'
+
+        app = TestApp(Pecan(RootController()))
+
+        r = app.get('/some-path/')
+        assert r.status_int == 200
+        assert r.body == b_('Hello, World!')
+
+        r = app.get('/some_path/', expect_errors=True)
+        assert r.status_int == 404
+
+    def test_manual_route(self):
+
+        class SubController(object):
+
+            @expose(route='some-path')
+            def some_path(self):
+                return 'Hello, World!'
+
+        class RootController(object):
+            pass
+
+        route(RootController, 'some-controller', SubController())
+
+        app = TestApp(Pecan(RootController()))
+
+        r = app.get('/some-controller/some-path/')
+        assert r.status_int == 200
+        assert r.body == b_('Hello, World!')
+
+        r = app.get('/some-controller/some_path/', expect_errors=True)
+        assert r.status_int == 404
+
+    def test_manual_route_conflict(self):
+
+        class SubController(object):
+            pass
+
+        class RootController(object):
+
+            @expose()
+            def hello(self):
+                return 'Hello, World!'
+
+        self.assertRaises(
+            RuntimeError,
+            route,
+            RootController,
+            'hello',
+            SubController()
+        )
+
+    def test_custom_route_on_index(self):
+
+        class RootController(object):
+
+            @expose(route='some-path')
+            def index(self):
+                return 'Hello, World!'
+
+        app = TestApp(Pecan(RootController()))
+
+        r = app.get('/some-path/')
+        assert r.status_int == 200
+        assert r.body == b_('Hello, World!')
+
+        r = app.get('/')
+        assert r.status_int == 200
+        assert r.body == b_('Hello, World!')
+
+        r = app.get('/index/', expect_errors=True)
+        assert r.status_int == 404
+
+    def test_custom_route_with_attribute_conflict(self):
+
+        class RootController(object):
+
+            @expose(route='mock')
+            def greet(self):
+                return 'Hello, World!'
+
+            @expose()
+            def mock(self):
+                return 'You are not worthy!'
+
+        app = TestApp(Pecan(RootController()))
+
+        self.assertRaises(
+            RuntimeError,
+            app.get,
+            '/mock/'
+        )
+
+    def test_conflicting_custom_routes(self):
+
+        class RootController(object):
+
+            @expose(route='testing')
+            def foo(self):
+                return 'Foo!'
+
+            @expose(route='testing')
+            def bar(self):
+                return 'Bar!'
+
+        app = TestApp(Pecan(RootController()))
+
+        self.assertRaises(
+            RuntimeError,
+            app.get,
+            '/testing/'
+        )
+
+    def test_conflicting_custom_routes_in_subclass(self):
+
+        class BaseController(object):
+
+            @expose(route='testing')
+            def foo(self):
+                return request.path
+
+        class ChildController(BaseController):
+            pass
+
+        class RootController(BaseController):
+            child = ChildController()
+
+        app = TestApp(Pecan(RootController()))
+
+        r = app.get('/testing/')
+        assert r.body == b_('/testing/')
+
+        r = app.get('/child/testing/')
+        assert r.body == b_('/child/testing/')
+
+    def test_custom_route_prohibited_on_lookup(self):
+        try:
+            class RootController(object):
+
+                @expose(route='some-path')
+                def _lookup(self):
+                    return 'Hello, World!'
+        except ValueError:
+            pass
+        else:
+            raise AssertionError(
+                '_lookup cannot be used with a custom path segment'
+            )
+
+    def test_custom_route_prohibited_on_default(self):
+        try:
+            class RootController(object):
+
+                @expose(route='some-path')
+                def _default(self):
+                    return 'Hello, World!'
+        except ValueError:
+            pass
+        else:
+            raise AssertionError(
+                '_default cannot be used with a custom path segment'
+            )
+
+    def test_custom_route_prohibited_on_route(self):
+        try:
+            class RootController(object):
+
+                @expose(route='some-path')
+                def _route(self):
+                    return 'Hello, World!'
+        except ValueError:
+            pass
+        else:
+            raise AssertionError(
+                '_route cannot be used with a custom path segment'
+            )
+
+    def test_custom_route_with_generic_controllers(self):
+
+        class RootController(object):
+
+            @expose(route='some-path', generic=True)
+            def foo(self):
+                return 'Hello, World!'
+
+            @foo.when(method='POST')
+            def handle_post(self):
+                return 'POST!'
+
+        app = TestApp(Pecan(RootController()))
+
+        r = app.get('/some-path/')
+        assert r.status_int == 200
+        assert r.body == b_('Hello, World!')
+
+        r = app.get('/foo/', expect_errors=True)
+        assert r.status_int == 404
+
+        r = app.post('/some-path/')
+        assert r.status_int == 200
+        assert r.body == b_('POST!')
+
+        r = app.post('/foo/', expect_errors=True)
+        assert r.status_int == 404
+
+    def test_custom_route_prohibited_on_generic_controllers(self):
+        try:
+            class RootController(object):
+
+                @expose(generic=True)
+                def foo(self):
+                    return 'Hello, World!'
+
+                @foo.when(method='POST', route='some-path')
+                def handle_post(self):
+                    return 'POST!'
+        except ValueError:
+            pass
+        else:
+            raise AssertionError(
+                'generic controllers cannot be used with a custom path segment'
+            )
+
+    def test_invalid_route_arguments(self):
+        class C(object):
+
+            def secret(self):
+                return {}
+
+        self.assertRaises(TypeError, route)
+        self.assertRaises(TypeError, route, 'some-path', lambda x: x)
+        self.assertRaises(TypeError, route, 'some-path', C.secret)
+        self.assertRaises(TypeError, route, C, {}, C())
+
+        for path in (
+            'VARIED-case-PATH',
+            'this,custom,path',
+            '123-path',
+            'path(with-parens)',
+            'path;with;semicolons',
+            'path:with:colons',
+            'v2.0',
+            '~username',
+            'somepath!',
+            'four*four',
+            'one+two',
+            '@twitterhandle',
+            'package=pecan'
+        ):
+            handler = C()
+            route(C, path, handler)
+            assert getattr(C, path, handler)
+
+        self.assertRaises(ValueError, route, C, '/path/', C())
+        self.assertRaises(ValueError, route, C, '.', C())
+        self.assertRaises(ValueError, route, C, '..', C())
+        self.assertRaises(ValueError, route, C, 'path?', C())
+        self.assertRaises(ValueError, route, C, 'percent%20encoded', C())
