@@ -1,8 +1,13 @@
+import inspect
+import operator
+
 from webtest import TestApp
+from six import PY3
 from six import b as b_
+from six import u as u_
 from six.moves import cStringIO as StringIO
 
-from pecan import make_app, expose, redirect, abort
+from pecan import make_app, expose, redirect, abort, rest, Request, Response
 from pecan.hooks import (
     PecanHook, TransactionHook, HookController, RequestViewerHook
 )
@@ -10,6 +15,9 @@ from pecan.configuration import Config
 from pecan.decorators import transactional, after_commit, after_rollback
 from pecan.tests import PecanTestCase
 
+# The `inspect.Arguments` namedtuple is different between PY2/3
+kwargs = operator.attrgetter('varkw' if PY3 else 'keywords')
+
 
 class TestHooks(PecanTestCase):
 
@@ -133,6 +141,33 @@ class TestHooks(PecanTestCase):
         assert run_hook[0] == 'on_route'
         assert run_hook[1] == 'error'
 
+    def test_on_error_response_hook(self):
+        run_hook = []
+
+        class RootController(object):
+            @expose()
+            def causeerror(self):
+                return [][1]
+
+        class ErrorHook(PecanHook):
+            def on_error(self, state, e):
+                run_hook.append('error')
+
+                r = Response()
+                r.text = u_('on_error')
+
+                return r
+
+        app = TestApp(make_app(RootController(), hooks=[
+            ErrorHook()
+        ]))
+
+        response = app.get('/causeerror')
+
+        assert len(run_hook) == 1
+        assert run_hook[0] == 'error'
+        assert response.text == 'on_error'
+
     def test_prioritized_hooks(self):
         run_hook = []
 
@@ -309,6 +344,436 @@ class TestHooks(PecanTestCase):
         assert run_hook[4] == 'after1'
         assert run_hook[5] == 'after2'
 
+    def test_mixin_hooks(self):
+        run_hook = []
+
+        class HelperHook(PecanHook):
+            priority = 2
+
+            def before(self, state):
+                run_hook.append('helper - before hook')
+
+        # we'll use the same hook instance to avoid duplicate calls
+        helper_hook = HelperHook()
+
+        class LastHook(PecanHook):
+            priority = 200
+
+            def before(self, state):
+                run_hook.append('last - before hook')
+
+        class SimpleHook(PecanHook):
+            priority = 1
+
+            def before(self, state):
+                run_hook.append('simple - before hook')
+
+        class HelperMixin(object):
+            __hooks__ = [helper_hook]
+
+        class LastMixin(object):
+            __hooks__ = [LastHook()]
+
+        class SubController(HookController, HelperMixin):
+            __hooks__ = [LastHook()]
+
+            @expose()
+            def index(self):
+                return "This is sub controller!"
+
+        class RootController(HookController, LastMixin):
+            __hooks__ = [SimpleHook(), helper_hook]
+
+            @expose()
+            def index(self):
+                run_hook.append('inside')
+                return 'Hello, World!'
+
+            sub = SubController()
+
+        papp = make_app(RootController())
+        app = TestApp(papp)
+        response = app.get('/')
+        assert response.status_int == 200
+        assert response.body == b_('Hello, World!')
+
+        assert len(run_hook) == 4
+        assert run_hook[0] == 'simple - before hook', run_hook[0]
+        assert run_hook[1] == 'helper - before hook', run_hook[1]
+        assert run_hook[2] == 'last - before hook', run_hook[2]
+        assert run_hook[3] == 'inside', run_hook[3]
+
+        run_hook = []
+        response = app.get('/sub/')
+        assert response.status_int == 200
+        assert response.body == b_('This is sub controller!')
+
+        assert len(run_hook) == 4, run_hook
+        assert run_hook[0] == 'simple - before hook', run_hook[0]
+        assert run_hook[1] == 'helper - before hook', run_hook[1]
+        assert run_hook[2] == 'last - before hook', run_hook[2]
+        # LastHook is invoked once again -
+        # for each different instance of the Hook in the two Controllers
+        assert run_hook[3] == 'last - before hook', run_hook[3]
+
+
+class TestStateAccess(PecanTestCase):
+
+    def setUp(self):
+        super(TestStateAccess, self).setUp()
+        self.args = None
+
+        class RootController(object):
+            @expose()
+            def index(self):
+                return 'Hello, World!'
+
+            @expose()
+            def greet(self, name):
+                return 'Hello, %s!' % name
+
+            @expose()
+            def greetmore(self, *args):
+                return 'Hello, %s!' % args[0]
+
+            @expose()
+            def kwargs(self, **kw):
+                return 'Hello, %s!' % kw['name']
+
+            @expose()
+            def mixed(self, first, second, *args):
+                return 'Mixed'
+
+        class SimpleHook(PecanHook):
+            def before(inself, state):
+                self.args = (state.controller, state.arguments)
+
+        self.root = RootController()
+        self.app = TestApp(make_app(self.root, hooks=[SimpleHook()]))
+
+    def test_no_args(self):
+        self.app.get('/')
+        assert self.args[0] == self.root.index
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert self.args[1].args == []
+        assert self.args[1].varargs == []
+        assert kwargs(self.args[1]) == {}
+
+    def test_single_arg(self):
+        self.app.get('/greet/joe')
+        assert self.args[0] == self.root.greet
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert self.args[1].args == ['joe']
+        assert self.args[1].varargs == []
+        assert kwargs(self.args[1]) == {}
+
+    def test_single_vararg(self):
+        self.app.get('/greetmore/joe')
+        assert self.args[0] == self.root.greetmore
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert self.args[1].args == []
+        assert self.args[1].varargs == ['joe']
+        assert kwargs(self.args[1]) == {}
+
+    def test_single_kw(self):
+        self.app.get('/kwargs/?name=joe')
+        assert self.args[0] == self.root.kwargs
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert self.args[1].args == []
+        assert self.args[1].varargs == []
+        assert kwargs(self.args[1]) == {'name': 'joe'}
+
+    def test_single_kw_post(self):
+        self.app.post('/kwargs/', params={'name': 'joe'})
+        assert self.args[0] == self.root.kwargs
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert self.args[1].args == []
+        assert self.args[1].varargs == []
+        assert kwargs(self.args[1]) == {'name': 'joe'}
+
+    def test_mixed_args(self):
+        self.app.get('/mixed/foo/bar/spam/eggs')
+        assert self.args[0] == self.root.mixed
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert self.args[1].args == ['foo', 'bar']
+        assert self.args[1].varargs == ['spam', 'eggs']
+
+
+class TestStateAccessWithoutThreadLocals(PecanTestCase):
+
+    def setUp(self):
+        super(TestStateAccessWithoutThreadLocals, self).setUp()
+        self.args = None
+
+        class RootController(object):
+            @expose()
+            def index(self, req, resp):
+                return 'Hello, World!'
+
+            @expose()
+            def greet(self, req, resp, name):
+                return 'Hello, %s!' % name
+
+            @expose()
+            def greetmore(self, req, resp, *args):
+                return 'Hello, %s!' % args[0]
+
+            @expose()
+            def kwargs(self, req, resp, **kw):
+                return 'Hello, %s!' % kw['name']
+
+            @expose()
+            def mixed(self, req, resp, first, second, *args):
+                return 'Mixed'
+
+        class SimpleHook(PecanHook):
+            def before(inself, state):
+                self.args = (state.controller, state.arguments)
+
+        self.root = RootController()
+        self.app = TestApp(make_app(
+            self.root,
+            hooks=[SimpleHook()],
+            use_context_locals=False
+        ))
+
+    def test_no_args(self):
+        self.app.get('/')
+        assert self.args[0] == self.root.index
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert len(self.args[1].args) == 2
+        assert isinstance(self.args[1].args[0], Request)
+        assert isinstance(self.args[1].args[1], Response)
+        assert self.args[1].varargs == []
+        assert kwargs(self.args[1]) == {}
+
+    def test_single_arg(self):
+        self.app.get('/greet/joe')
+        assert self.args[0] == self.root.greet
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert len(self.args[1].args) == 3
+        assert isinstance(self.args[1].args[0], Request)
+        assert isinstance(self.args[1].args[1], Response)
+        assert self.args[1].args[2] == 'joe'
+        assert self.args[1].varargs == []
+        assert kwargs(self.args[1]) == {}
+
+    def test_single_vararg(self):
+        self.app.get('/greetmore/joe')
+        assert self.args[0] == self.root.greetmore
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert len(self.args[1].args) == 2
+        assert isinstance(self.args[1].args[0], Request)
+        assert isinstance(self.args[1].args[1], Response)
+        assert self.args[1].varargs == ['joe']
+        assert kwargs(self.args[1]) == {}
+
+    def test_single_kw(self):
+        self.app.get('/kwargs/?name=joe')
+        assert self.args[0] == self.root.kwargs
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert len(self.args[1].args) == 2
+        assert isinstance(self.args[1].args[0], Request)
+        assert isinstance(self.args[1].args[1], Response)
+        assert self.args[1].varargs == []
+        assert kwargs(self.args[1]) == {'name': 'joe'}
+
+    def test_single_kw_post(self):
+        self.app.post('/kwargs/', params={'name': 'joe'})
+        assert self.args[0] == self.root.kwargs
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert len(self.args[1].args) == 2
+        assert isinstance(self.args[1].args[0], Request)
+        assert isinstance(self.args[1].args[1], Response)
+        assert self.args[1].varargs == []
+        assert kwargs(self.args[1]) == {'name': 'joe'}
+
+    def test_mixed_args(self):
+        self.app.get('/mixed/foo/bar/spam/eggs')
+        assert self.args[0] == self.root.mixed
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert len(self.args[1].args) == 4
+        assert isinstance(self.args[1].args[0], Request)
+        assert isinstance(self.args[1].args[1], Response)
+        assert self.args[1].args[2:] == ['foo', 'bar']
+        assert self.args[1].varargs == ['spam', 'eggs']
+
+
+class TestRestControllerStateAccess(PecanTestCase):
+
+    def setUp(self):
+        super(TestRestControllerStateAccess, self).setUp()
+        self.args = None
+
+        class RootController(rest.RestController):
+
+            @expose()
+            def _default(self, _id, *args, **kw):
+                return 'Default'
+
+            @expose()
+            def get_all(self, **kw):
+                return 'All'
+
+            @expose()
+            def get_one(self, _id, *args, **kw):
+                return 'One'
+
+            @expose()
+            def post(self, *args, **kw):
+                return 'POST'
+
+            @expose()
+            def put(self, _id, *args, **kw):
+                return 'PUT'
+
+            @expose()
+            def delete(self, _id, *args, **kw):
+                return 'DELETE'
+
+        class SimpleHook(PecanHook):
+            def before(inself, state):
+                self.args = (state.controller, state.arguments)
+
+        self.root = RootController()
+        self.app = TestApp(make_app(self.root, hooks=[SimpleHook()]))
+
+    def test_get_all(self):
+        self.app.get('/')
+        assert self.args[0] == self.root.get_all
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert self.args[1].args == []
+        assert self.args[1].varargs == []
+        assert kwargs(self.args[1]) == {}
+
+    def test_get_all_with_kwargs(self):
+        self.app.get('/?foo=bar')
+        assert self.args[0] == self.root.get_all
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert self.args[1].args == []
+        assert self.args[1].varargs == []
+        assert kwargs(self.args[1]) == {'foo': 'bar'}
+
+    def test_get_one(self):
+        self.app.get('/1')
+        assert self.args[0] == self.root.get_one
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert self.args[1].args == ['1']
+        assert self.args[1].varargs == []
+        assert kwargs(self.args[1]) == {}
+
+    def test_get_one_with_varargs(self):
+        self.app.get('/1/2/3')
+        assert self.args[0] == self.root.get_one
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert self.args[1].args == ['1']
+        assert self.args[1].varargs == ['2', '3']
+        assert kwargs(self.args[1]) == {}
+
+    def test_get_one_with_kwargs(self):
+        self.app.get('/1?foo=bar')
+        assert self.args[0] == self.root.get_one
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert self.args[1].args == ['1']
+        assert self.args[1].varargs == []
+        assert kwargs(self.args[1]) == {'foo': 'bar'}
+
+    def test_post(self):
+        self.app.post('/')
+        assert self.args[0] == self.root.post
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert self.args[1].args == []
+        assert self.args[1].varargs == []
+        assert kwargs(self.args[1]) == {}
+
+    def test_post_with_varargs(self):
+        self.app.post('/foo/bar')
+        assert self.args[0] == self.root.post
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert self.args[1].args == []
+        assert self.args[1].varargs == ['foo', 'bar']
+        assert kwargs(self.args[1]) == {}
+
+    def test_post_with_kwargs(self):
+        self.app.post('/', params={'foo': 'bar'})
+        assert self.args[0] == self.root.post
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert self.args[1].args == []
+        assert self.args[1].varargs == []
+        assert kwargs(self.args[1]) == {'foo': 'bar'}
+
+    def test_put(self):
+        self.app.put('/1')
+        assert self.args[0] == self.root.put
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert self.args[1].args == ['1']
+        assert self.args[1].varargs == []
+        assert kwargs(self.args[1]) == {}
+
+    def test_put_with_method_argument(self):
+        self.app.post('/1?_method=put')
+        assert self.args[0] == self.root.put
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert self.args[1].args == ['1']
+        assert self.args[1].varargs == []
+        assert kwargs(self.args[1]) == {'_method': 'put'}
+
+    def test_put_with_varargs(self):
+        self.app.put('/1/2/3')
+        assert self.args[0] == self.root.put
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert self.args[1].args == ['1']
+        assert self.args[1].varargs == ['2', '3']
+        assert kwargs(self.args[1]) == {}
+
+    def test_put_with_kwargs(self):
+        self.app.put('/1?foo=bar')
+        assert self.args[0] == self.root.put
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert self.args[1].args == ['1']
+        assert self.args[1].varargs == []
+        assert kwargs(self.args[1]) == {'foo': 'bar'}
+
+    def test_delete(self):
+        self.app.delete('/1')
+        assert self.args[0] == self.root.delete
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert self.args[1].args == ['1']
+        assert self.args[1].varargs == []
+        assert kwargs(self.args[1]) == {}
+
+    def test_delete_with_method_argument(self):
+        self.app.post('/1?_method=delete')
+        assert self.args[0] == self.root.delete
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert self.args[1].args == ['1']
+        assert self.args[1].varargs == []
+        assert kwargs(self.args[1]) == {'_method': 'delete'}
+
+    def test_delete_with_varargs(self):
+        self.app.delete('/1/2/3')
+        assert self.args[0] == self.root.delete
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert self.args[1].args == ['1']
+        assert self.args[1].varargs == ['2', '3']
+        assert kwargs(self.args[1]) == {}
+
+    def test_delete_with_kwargs(self):
+        self.app.delete('/1?foo=bar')
+        assert self.args[0] == self.root.delete
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert self.args[1].args == ['1']
+        assert self.args[1].varargs == []
+        assert kwargs(self.args[1]) == {'foo': 'bar'}
+
+    def test_post_with_invalid_method_kwarg(self):
+        self.app.post('/1?_method=invalid')
+        assert self.args[0] == self.root._default
+        assert isinstance(self.args[1], inspect.Arguments)
+        assert self.args[1].args == ['1']
+        assert self.args[1].varargs == []
+        assert kwargs(self.args[1]) == {'_method': 'invalid'}
+
 
 class TestTransactionHook(PecanTestCase):
     def test_transaction_hook(self):
@@ -1017,21 +1482,6 @@ class TestTransactionHook(PecanTestCase):
 
 class TestRequestViewerHook(PecanTestCase):
 
-    def test_hook_from_config(self):
-        from pecan.configuration import _runtime_conf as conf
-        conf['requestviewer'] = {
-            'blacklist': ['/favicon.ico']
-        }
-
-        class RootController(object):
-            pass
-
-        app = make_app(RootController())
-        while hasattr(app, 'application'):
-            app = app.application
-        del conf.__values__['requestviewer']
-        assert app.hooks
-
     def test_basic_single_default_hook(self):
 
         _stdout = StringIO()
@@ -1043,7 +1493,9 @@ class TestRequestViewerHook(PecanTestCase):
 
         app = TestApp(
             make_app(
-                RootController(), hooks=[RequestViewerHook(writer=_stdout)]
+                RootController(), hooks=lambda: [
+                    RequestViewerHook(writer=_stdout)
+                ]
             )
         )
         response = app.get('/')
@@ -1074,7 +1526,9 @@ class TestRequestViewerHook(PecanTestCase):
 
         app = TestApp(
             make_app(
-                RootController(), hooks=[RequestViewerHook(writer=_stdout)]
+                RootController(), hooks=lambda: [
+                    RequestViewerHook(writer=_stdout)
+                ]
             )
         )
         response = app.get('/404', expect_errors=True)
@@ -1104,7 +1558,7 @@ class TestRequestViewerHook(PecanTestCase):
         app = TestApp(
             make_app(
                 RootController(),
-                hooks=[
+                hooks=lambda: [
                     RequestViewerHook(
                         config={'items': ['path']}, writer=_stdout
                     )
@@ -1139,7 +1593,7 @@ class TestRequestViewerHook(PecanTestCase):
         app = TestApp(
             make_app(
                 RootController(),
-                hooks=[
+                hooks=lambda: [
                     RequestViewerHook(
                         config={'blacklist': ['/']}, writer=_stdout
                     )
@@ -1166,7 +1620,7 @@ class TestRequestViewerHook(PecanTestCase):
         app = TestApp(
             make_app(
                 RootController(),
-                hooks=[
+                hooks=lambda: [
                     RequestViewerHook(
                         config={'items': ['date']}, writer=_stdout
                     )
@@ -1202,3 +1656,56 @@ class TestRequestViewerHook(PecanTestCase):
         viewer = RequestViewerHook(conf)
 
         assert viewer.items == ['url']
+
+
+class TestRestControllerWithHooks(PecanTestCase):
+
+    def test_restcontroller_with_hooks(self):
+
+        class SomeHook(PecanHook):
+
+            def before(self, state):
+                state.response.headers['X-Testing'] = 'XYZ'
+
+        class BaseController(rest.RestController):
+
+            @expose()
+            def delete(self, _id):
+                return 'Deleting %s' % _id
+
+        class RootController(BaseController, HookController):
+
+            __hooks__ = [SomeHook()]
+
+            @expose()
+            def get_all(self):
+                return 'Hello, World!'
+
+            @staticmethod
+            def static(cls):
+                return 'static'
+
+            @property
+            def foo(self):
+                return 'bar'
+
+            def testing123(self):
+                return 'bar'
+
+            unhashable = [1, 'two', 3]
+
+        app = TestApp(
+            make_app(
+                RootController()
+            )
+        )
+
+        response = app.get('/')
+        assert response.status_int == 200
+        assert response.body == b_('Hello, World!')
+        assert response.headers['X-Testing'] == 'XYZ'
+
+        response = app.delete('/100/')
+        assert response.status_int == 200
+        assert response.body == b_('Deleting 100')
+        assert response.headers['X-Testing'] == 'XYZ'
